/*
 *    rttytx.c  --  RTTY modulator
 *
 *    Copyright (C) 2001, 2002, 2003
 *      Tomi Manninen (oh2bns@sral.fi)
 *
 *    This file is part of gMFSK.
 *
 *    gMFSK is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    gMFSK is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with gMFSK; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "trx.h"
#include "rtty.h"
#include "snd.h"
#include "baudot.h"
#include "rttypar.h"

static inline double nco(struct rtty *s, double freq)
{
	s->phaseacc += 2.0 * M_PI * freq / SampleRate;

	if (s->phaseacc > M_PI)
		s->phaseacc -= 2.0 * M_PI;

	return cos(s->phaseacc);
}

static void send_symbol(struct trx *trx, int symbol)
{
	struct rtty *s = (struct rtty *) trx->modem;
	double freq;
	int i, rev;

	rev = (trx->reverse != 0) ^ (s->reverse != 0);

	if (rev)
		symbol = !symbol;

#if WANT_FREQLOCK
	freq = trx->txfrequency;
#else
	freq = trx->frequency;
#endif
	if (symbol)
		freq -= s->shift / 2.0;
	else
		freq += s->shift / 2.0;

	freq += trx->txoffset;

	for (i = 0; i < s->symbollen; i++)
		trx->outbuf[i] = nco(s, freq);

	sound_write(trx->outbuf, s->symbollen);
}

static void send_stop(struct trx *trx)
{
	struct rtty *s = (struct rtty *) trx->modem;
	double freq;
	int i, rev;

	rev = (trx->reverse != 0) ^ (s->reverse != 0);

	if (rev)
		freq = trx->frequency + s->shift / 2.0;
	else
		freq = trx->frequency - s->shift / 2.0;

	freq += trx->txoffset;

	for (i = 0; i < s->stoplen; i++)
		trx->outbuf[i] = nco(s, freq);

	sound_write(trx->outbuf, s->stoplen);
}

static void send_char(struct trx *trx, int c)
{
	struct rtty *s = (struct rtty *) trx->modem;
	int i, j;

	if (s->nbits == 5) {
		if (c == BAUDOT_LETS)
			c = 0x1F;
		if (c == BAUDOT_FIGS)
			c = 0x1B;
	}

	/* start bit */
	send_symbol(trx, 0);

	/* data bits */
	for (i = 0; i < s->nbits; i++) {
		j = (s->msb) ? (s->nbits - 1 - i) : i;
		send_symbol(trx, (c >> j) & 1);
	}

	/* parity bit */
	if (s->parity != PARITY_NONE)
		send_symbol(trx, rtty_parity(c, s->nbits, s->parity));

	/* stop bit(s) */
	send_stop(trx);

	if (s->nbits == 5)
	  c = baudot_dec(&s->txmode, c);
// following line does not work correctly	  
//		c = baudot_dec(&s->rxmode, c);

#if WANT_TTY_MOD
	if (c != 0 && c != '\r')
#else
	if (c != 0)
#endif
		trx_put_echo_char(c);
}

static void send_idle(struct trx *trx)
{
	struct rtty *s = (struct rtty *) trx->modem;

	if (s->nbits == 5)
		send_char(trx, BAUDOT_LETS);
	else
		send_char(trx, 0);
}

#if WANT_TTY_MOD
int line_char_count = 0;
#endif

int rtty_txprocess(struct trx *trx)
{
	struct rtty *s = (struct rtty *) trx->modem;
	int c;

	if (trx->tune) {
		trx->tune = 0;
		send_symbol(trx, 1);
		return 0;
	}

	if (s->preamble > 0) {
		send_symbol(trx, 1);
		s->preamble--;
		return 0;
	}

	c = trx_get_tx_char();

	/* TX buffer empty */
	if (c == -1) {
		/* stop if requested to... */

#if WANT_TTY_MOD
		if (trx->stopflag) {
			line_char_count = 0;
			send_char(trx,0x08);
			send_char(trx,0x08);
			send_char(trx,0x02);
			return -1;
		}
#else		
		if (trx->stopflag)
			return -1;
#endif

		/* send idle character */
		send_idle(trx);
		s->txmode = BAUDOT_LETS;
		return 0;
	}

	if (s->nbits != 5) {
		send_char(trx, c);
		return 0;
	}

#if WANT_TTY_MOD
  if (isalpha(c) || isdigit(c) || isblank(c) || ispunct(c)) {
    ++line_char_count;
	}

	if ( (c != '\n' && c != '\r') && 
			 (line_char_count == 72 || (line_char_count > 66 && c == ' '))) {
		line_char_count = 0;
		send_char(trx,0x08); // CR-CR-LF triplet
		send_char(trx,0x08);
		send_char(trx,0x02);
		if (c == ' ')
			return 0;
	}
	if (c == '\r') {
		line_char_count = 0;
		send_char(trx, 0x08); // CR-CR pair
		send_char(trx, 0x08);
		return 0;
	}
	if (c == '\n') {
		line_char_count = 0;
		send_char(trx, 0x08); // CR-CR-LF triplet
		send_char(trx, 0x08);
		send_char(trx, 0x02);
		return 0;
	}
#endif

/* unshift-on-space */
	if (c == ' ') {
		send_char(trx, BAUDOT_LETS);
		send_char(trx, 0x04); // coded value for a space
		s->txmode = BAUDOT_LETS;
		return 0;
	}

	if ((c = baudot_enc(c)) < 0)
		return 0;

	/* switch case if necessary */
	if ((c & s->txmode) == 0) {
		if (s->txmode == BAUDOT_FIGS) {
			send_char(trx, BAUDOT_LETS);
			s->txmode = BAUDOT_LETS;
		} else {
			send_char(trx, BAUDOT_FIGS);
			s->txmode = BAUDOT_FIGS;
		}
	}

	send_char(trx, c & 0x1F);

	return 0;
}
